/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "hra.h"
#include "kocka.h"

#include <nnhk.h>
#include <troj3d.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#define VELKOSTH_KOCKY 16
#define VELKOSTH_BLOKU 16

#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))

#define VIDENIE_CSHANKOV 1

/* Ryychlosti v m/s. */
#define RYYCHLOST_CHUODZE      4.317
#define RYYCHLOST_SHPRINTU     5.612
#define RYYCHLOST_PLIIZHENIA   4.3
#define RYYCHLOST_LETU         11.0
#define RYYCHLOST_LET_SHPRINTU 22.0
/* Padanie: Kazhdyy tik -0.08 m/s, a potom * 0.98. */

static void dajTrojuhZKocky(TROJUH *trojuh, int index, int x, int y, int z)
{
    int i;
    int ib = index * 9;
    int v = VELKOSTH_KOCKY / 2;

    for (i = 0; i < 3; i++)
    {
        trojuh->vrcholy[i].x = kockaTrojuhy[ib++] * v + x * VELKOSTH_KOCKY;
        trojuh->vrcholy[i].y = kockaTrojuhy[ib++] * v + y * VELKOSTH_KOCKY;
        trojuh->vrcholy[i].z = kockaTrojuhy[ib++] * v + z * VELKOSTH_KOCKY;
    }
}

/* Pred - 0, Vrch - 1, Lhavaa - 2... potom oproti. */

static void dajStranu(TROJUH *trojuhy, int strana, int x, int y, int z,
                      unsigned long farba)
{
    dajTrojuhZKocky(trojuhy, strana * 2, x, y, z);
    dajTrojuhZKocky(trojuhy + 1, strana * 2 + 1, x, y, z);
    trojuhy[0].farba = farba;
    trojuhy[1].farba = farba;
}

static unsigned long farby[6] = {0x000000ff, 0x0000ff00, 0x00ff0000,
                                 0x007f0000, 0x00007f00, 0x0000007f};

static void cshankGenBloky(CSHANK *cshank)
{
    int x, y, z;

    memset(cshank->bloky, 0, sizeof(cshank->bloky));

    for (x = 0; x < CSHANK_SHIIRKA; x++)
    {
        for (z = 0; z < CSHANK_DLLZHKA; z++)
        {
            for (y = 0; y < 3 + rand() % 2; y++)
            {
                cshank->bloky[x][y][z].vidithelhnyy = 1;
                cshank->bloky[x][y][z].farba = farby[rand() % 6];
            }
        }
    }
}

static int cshankKockaNa(CSHANK *cshank, int x, int y, int z)
{
    if (x < 0 || x >= CSHANK_SHIIRKA
        || y < 0 || y >= CSHANK_VYYSHKA
        || z < 0 || z >= CSHANK_DLLZHKA)
    {
        return (0);
    }

    return (cshank->bloky[x][y][z].vidithelhnyy);
}

static int cshankDavajKocku(CSHANK *cshank,
                            TROJUH *trojuhy,
                            int x, int y, int z,
                            unsigned long farba)
{
    int gene = 0;

    if (cshankKockaNa(cshank, x, y, z - 1) == 0)
    {
        dajStranu(trojuhy + gene, 0, x, y, z, farba);
        gene += 2;
    }

    if (cshankKockaNa(cshank, x, y + 1, z) == 0)
    {
        dajStranu(trojuhy + gene, 1, x, y, z, farba);
        gene += 2;
    }

    if (cshankKockaNa(cshank, x - 1, y, z) == 0)
    {
        dajStranu(trojuhy + gene, 2, x, y, z, farba);
        gene += 2;
    }

    if (cshankKockaNa(cshank, x + 1, y, z) == 0)
    {
        dajStranu(trojuhy + gene, 3, x, y, z, farba);
        gene += 2;
    }

    if (cshankKockaNa(cshank, x, y - 1, z) == 0)
    {
        dajStranu(trojuhy + gene, 4, x, y, z, farba);
        gene += 2;
    }

    if (cshankKockaNa(cshank, x, y, z + 1) == 0)
    {
        dajStranu(trojuhy + gene, 5, x, y, z, farba);
        gene += 2;
    }

    return (gene);
}

static void cshankGenKocky(CSHANK *cshank)
{
    int x, y, z;

    cshank->n_trojuhy = 0;

    for (x = 0; x < CSHANK_SHIIRKA; x++)
    {
        for (y = 0; y < CSHANK_VYYSHKA; y++)
        {
            for (z = 0; z < CSHANK_DLLZHKA; z++)
            {
                if (cshank->bloky[x][y][z].vidithelhnyy == 0) continue;

                cshank->n_trojuhy += cshankDavajKocku(cshank,
                                                      cshank->trojuhy
                                                      + (cshank->n_trojuhy),
                                                      x, y, z,
                                                      cshank->
                                                      bloky[x][y][z].farba);
            }
        }
    }
}

static void posunhTrojuhy(TROJUH *trojuhy, int n_trojuhy,
                          int x, int y, int z)
{
    int i, j;

    for (i = 0; i < n_trojuhy; i++)
    {
        for (j = 0; j < 3; j++)
        {
            trojuhy[i].vrcholy[j].x += x;
            trojuhy[i].vrcholy[j].y += y;
            trojuhy[i].vrcholy[j].z += z;
        }
    }
}

#define SHIIRKA_HRAACSHA 0.60
#define VYYSHKA_HRAACSHA 1.70

static void hraacshNastavKameru(HRAACSH *hraacsh, KAMERA *kamera)
{
    kamera->poz.x = hraacsh->poz.x * VELKOSTH_BLOKU;
    kamera->poz.y = (hraacsh->poz.y + VYYSHKA_HRAACSHA) * VELKOSTH_BLOKU;
    kamera->poz.z = hraacsh->poz.z * VELKOSTH_BLOKU;

    kamera->rot.x = hraacsh->rot.x;
    kamera->rot.y = hraacsh->rot.y;
    kamera->rot.z = hraacsh->rot.z;
}

static void *bezpecshnyyMalloc(size_t v)
{
    void *r = malloc(v);

    if (r == NULL)
    {
        printf("Zlyhanie pri alokaacii %u bajtov\n", v);

        exit(-1);
    }

    return (r);
}

static double hraDajDivZ(HRA *hra)
{
    double zornee_pole_y = hra->zornee_pole * (2 * acos(-1)) / 360.0;
    double pomer = hra->shiirka / (hra->vyyshka);
    zornee_pole_y *= max(1, min(1.3999999999, sqrt(1.6 / pomer)));

    return (hra->vyyshka / 2 * (1 / tan(zornee_pole_y / 2)));
}

static void hraacshCshankIPoz(HRAACSH *hraacsh, IVEKT *ivekt)
{
    ivekt->x = hraacsh->poz.x / CSHANK_SHIIRKA;
    ivekt->y = hraacsh->poz.y / CSHANK_VYYSHKA;
    ivekt->z = hraacsh->poz.z / CSHANK_DLLZHKA;
}

static void cshankyNasilu(CSHANK ***cshankZoznam, IVEKT *iPoz)
{
    int i, j;

    for (i = 0; i < 3; i++)
    {
        for (j = 0; j < 3; j++)
        {
            cshankGenBloky(cshankZoznam[i][j]);
            cshankGenKocky(cshankZoznam[i][j]);
            posunhTrojuhy(cshankZoznam[i][j]->trojuhy,
                          cshankZoznam[i][j]->n_trojuhy,
                          (iPoz->x + i) * CSHANK_SHIIRKA * VELKOSTH_BLOKU,
                          0,
                          (iPoz->z + j) * CSHANK_DLLZHKA * VELKOSTH_BLOKU);
        }
    }
}

void hraInit(HRA *hra, NNHK_OKNO *okno)
{
    int i, j;

    hra->okno = okno;

    hra->shiirka = 640;
    hra->vyyshka = 480;
    hra->zornee_pole = 70;

    hra->kamera = bezpecshnyyMalloc(sizeof(KAMERA));
    {
        VEKTOR poz = {105, 53, -200};
        VEKTOR rot = {0, 0, 0};
        VEKTOR div = {-640 / 2, -480 / 2, hraDajDivZ(hra)};
        kameraInic(hra->kamera, &poz, &rot, &div);
    }

    hra->cshankZoznam = bezpecshnyyMalloc(sizeof(CSHANK **) * 3);
    for (i = 0; i < 3; i++)
    {
        hra->cshankZoznam[i] = bezpecshnyyMalloc(sizeof(CSHANK *) * 3);
        for (j = 0; j < 3; j++)
        {
            hra->cshankZoznam[i][j] = bezpecshnyyMalloc(sizeof(CSHANK));
            memset(hra->cshankZoznam[i][j], 0, sizeof(CSHANK));
        }
    }

    hra->hraacsh = bezpecshnyyMalloc(sizeof(HRAACSH));
    hra->hraacsh->poz.x = 997;
    hra->hraacsh->poz.y = 3;
    hra->hraacsh->poz.z = 999;

    hra->hraacsh->rot.x = 0;
    hra->hraacsh->rot.y = 0;
    hra->hraacsh->rot.z = 0;

    hra->v_hre = 0;

    hraacshCshankIPoz(hra->hraacsh, &(hra->cshankyIPoz));
    hra->cshankyIPoz.x -= VIDENIE_CSHANKOV;
    hra->cshankyIPoz.y -= VIDENIE_CSHANKOV;
    hra->cshankyIPoz.z -= VIDENIE_CSHANKOV;

    cshankyNasilu(hra->cshankZoznam, &(hra->cshankyIPoz));
}

void hraKoniec(HRA *hra)
{
    int i, j;

    free(hra->hraacsh);

    for (i = 0; i < 3; i++)
    {
        for (j = 0; j < 3; j++)
        {
            free(hra->cshankZoznam[i][j]);
        }
        free(hra->cshankZoznam[i]);
    }
    free(hra->cshankZoznam);

    free(hra->kamera);
}

void hraKresli(HRA *hra, void *sniimka)
{
    int i, j;

    hraacshNastavKameru(hra->hraacsh, hra->kamera);

    troj3dSniimka(sniimka, hra->shiirka, hra->vyyshka);

    for (i = 0; i < 3; i++)
    {
        for (j = 0; j < 3; j++)
        {
            zobrazTrojuhy(hra->kamera, sniimka, hra->shiirka, hra->vyyshka,
                          hra->cshankZoznam[i][j]->trojuhy,
                          hra->cshankZoznam[i][j]->n_trojuhy);
        }
    }
}

static char tlacshidlaa[0xff];

static void test(HRA *hra, const double dt)
{
    if (tlacshidlaa[VT_VLHAVO])
    {
        hra->hraacsh->v.x -= RYYCHLOST_CHUODZE * dt;
    }

    if (tlacshidlaa[VT_VPRAVO])
    {
        hra->hraacsh->v.x += RYYCHLOST_CHUODZE * dt;
    }

    if (tlacshidlaa[VT_HORE])
    {
        hra->hraacsh->v.z += RYYCHLOST_CHUODZE * dt;
    }

    if (tlacshidlaa[VT_DOLE])
    {
        hra->hraacsh->v.z -= RYYCHLOST_CHUODZE * dt;
    }

    if (tlacshidlaa[VT_O])
    {
        hra->hraacsh->poz.y -= 1;
    }

    if (tlacshidlaa[VT_P])
    {
        hra->hraacsh->poz.y += 1;
    }
}

static int spracujVstup(HRA *hra)
{
    UDALOSTH u;

    while (udalosthOdober(&u) != 0)
    {
        switch (u.typ) {
            case STLACSHENEE_TLACSHIDLO: {
                tlacshidlaa[u.daata.tlacshidlo] = 1;

                if (u.daata.tlacshidlo == VT_ESC)
                {
                    nnhk_OknoNastavChytenie(hra->okno, 0);
                }

                break;
            }

            case PUSTENEE_TLACSHIDLO: {
                tlacshidlaa[u.daata.tlacshidlo] = 0;

                break;
            }

            case MYSH_POHYB: {
                if (hra->v_hre)
                {
                    hra->hraacsh->rot.y += (u.daata.mysh_pohyb.rel.x
                                            * (2 * acos(-1)) / 360);
                    /* Odstranhuje zbytocshnee +360 a -360. */
                    if (hra->hraacsh->rot.y < -(acos(-1) * 2))
                    {
                        hra->hraacsh->rot.y += acos(-1) * 2;
                    }
                    if (hra->hraacsh->rot.y > (acos(-1) * 2))
                    {
                        hra->hraacsh->rot.y -= acos(-1) * 2;
                    }

                    hra->hraacsh->rot.x += (u.daata.mysh_pohyb.rel.y
                                            * (2 * acos(-1)) / 360);
                    /* Limituje otaacshanie okolo Xovej osi medzi 90 a -90. */
                    if (hra->hraacsh->rot.x < -(acos(-1) / 2))
                    {
                        hra->hraacsh->rot.x = -(acos(-1) / 2);
                    }
                    else if (hra->hraacsh->rot.x > (acos(-1) / 2))
                    {
                        hra->hraacsh->rot.x = acos(-1) / 2;
                    }
                }

                break;
            }

            case STRATA_DRZHANIA: {
                hra->v_hre = 0;

                break;
            }

            case NNHK_KONIEC: {
                return (1);
            }
        }
    }

    return (0);
}

int signum(const double h)
{
    return ((h > 0) - (h < 0));
}

void vektScshiitaj(VEKTOR *dest, VEKTOR *zdroj)
{
    dest->x += zdroj->x;
    dest->y += zdroj->y;
    dest->z += zdroj->z;
}

static void brzdi(double *d, const double zmena)
{
    if (*d > 0)
    {
        *d -= fabs(zmena);
        if (*d < 0) *d = 0;

        return;
    }

    if (*d < 0)
    {
        *d += fabs(zmena);
        if (*d > 0) *d = 0;
    }
}

typedef struct {
    VEKTOR min;
    VEKTOR max;
} KRABICA;

static void hraacshDajKrabicu(HRAACSH *hraacsh, KRABICA *krab)
{
    /* Hraacshovee XY suuradnice suu jeho Yovaa os. */
    krab->min.x = hraacsh->poz.x - SHIIRKA_HRAACSHA / 2;
    krab->min.y = hraacsh->poz.y;
    krab->min.z = hraacsh->poz.z - SHIIRKA_HRAACSHA / 2;

    krab->max.x = hraacsh->poz.x + SHIIRKA_HRAACSHA / 2;
    krab->max.y = hraacsh->poz.y + VYYSHKA_HRAACSHA;
    krab->max.z = hraacsh->poz.z + SHIIRKA_HRAACSHA / 2;
}

static void kockaDajKrabicu(VEKTOR *poz, KRABICA *krab)
{
    /* Kocka maa lhavyy prednyy roh na celyych suuradniciach. */
    krab->min.x = poz->x;
    krab->min.y = poz->y;
    krab->min.z = poz->z;

    krab->max.x = poz->x + 1;
    krab->max.y = poz->y + 1;
    krab->max.z = poz->z + 1;
}

static int krabiceKoliduj(KRABICA *krab1, KRABICA *krab2)
{
    if ((krab1->min.x <= krab2->max.x && krab1->max.x >= krab2->min.x)
        && (krab1->min.y <= krab2->max.y && krab1->max.y >= krab2->min.y)
        && (krab1->min.z <= krab2->max.z && krab1->max.z >= krab2->min.z))
    {
        return (1);
    }

    return (0);
}

static void hraacshRelPoz(HRAACSH *hraacsh, IVEKT *relPoz)
{
    relPoz->x = ((int)(hraacsh->poz.x)) % CSHANK_SHIIRKA;
    relPoz->y = ((int)(hraacsh->poz.y)) % CSHANK_VYYSHKA;
    relPoz->z = ((int)(hraacsh->poz.z)) % CSHANK_DLLZHKA;

    if (relPoz->x < 0) relPoz->x += CSHANK_SHIIRKA;
    if (relPoz->y < 0) relPoz->y += CSHANK_VYYSHKA;
    if (relPoz->z < 0) relPoz->z += CSHANK_DLLZHKA;
}

int hraTik(HRA *hra)
{
    int ret;
    const double dt = CSHAS_TIKU / 1000.0;

    ret = spracujVstup(hra);
    test(hra, dt);

    hra->hraacsh->poz.x += hra->hraacsh->v.x;
    hra->hraacsh->poz.y += hra->hraacsh->v.y;
    hra->hraacsh->poz.z += hra->hraacsh->v.z;

    brzdi(&(hra->hraacsh->v.x), RYYCHLOST_CHUODZE * dt);
    brzdi(&(hra->hraacsh->v.z), RYYCHLOST_CHUODZE * dt);

    {
        KRABICA h, k;
        VEKTOR vekt = {hra->hraacsh->poz.x,
                       hra->hraacsh->poz.y - 1,
                       hra->hraacsh->poz.z};
        IVEKT relPoz;

        hraacshRelPoz(hra->hraacsh, &relPoz);

        if (hra->cshankZoznam[1][1]->bloky[relPoz.x]
            [relPoz.y]
            [relPoz.z].vidithelhnyy)
        {
            hraacshDajKrabicu(hra->hraacsh, &h);
            kockaDajKrabicu(&vekt, &k);

            if (krabiceKoliduj(&h, &k))
            {
                hra->hraacsh->v.y = 0;
                hra->hraacsh->poz.y = (int)(hra->hraacsh->poz.y + 1);
            }
        }
    }

    /*hra->hraacsh->v.y -= 0.08;
    hra->hraacsh->v.y *= 0.98;*/

    return (ret);
}
