/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include "troj3d.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "kamera.h"
#include "vektor.h"

/* Postup:
 * Kamerovaa transformaacia (posun + otocshenie).
 * 3D rezanie.
 * Perspektiivne delenie.
 * (2D rezanie), rasterizaacia. */

#define max(a, b) (a > b ? a : b)
#define min(a, b) (a < b ? a : b)

void vrcholPrint(VRCHOL *v)
{
    printf("%f %f %f\n", v->x, v->y, v->z);
}

#define ROZX640 640
#define ROZY480 480

int hrany = 0;

static void perspektiivneDelenie(KAMERA *kamera, VRCHOL *vrchol)
{
    vrchol->x = vrchol->x * (kamera->div.z / (vrchol->z)) - kamera->div.x;
    vrchol->y = vrchol->y * (kamera->div.z / (vrchol->z)) - kamera->div.y;
}

static void projekcia(KAMERA *kamera, VRCHOL *vrchol, VRCHOL *novyyVrchol)
{
    double x, y, z;
    double dX, dY, dZ;

    x = vrchol->x - kamera->poz.x;
    y = vrchol->y - kamera->poz.y;
    z = vrchol->z - kamera->poz.z;

    dX = kamera->kos.y*(kamera->siin.z*y+kamera->kos.z*x)-kamera->siin.y*z;
    dY = (kamera->siin.x*(kamera->kos.y*z+kamera->siin.y
                         *(kamera->siin.z*y+kamera->kos.z*x))
          +kamera->kos.x*(kamera->kos.z*y-kamera->siin.z*x));
    dZ = (kamera->kos.x*(kamera->kos.y*z+kamera->siin.y
                        *(kamera->siin.z*y+kamera->kos.z*x))
          -kamera->siin.x*(kamera->kos.z*y-kamera->siin.z*x));

    novyyVrchol->x = dX;
    novyyVrchol->y = dY;
    novyyVrchol->z = dZ;
}

/* Makraa suu ryychlejshie nezh funkcie a o dosth. */
#define KLAMP(vstup) ((vstup) < 0 ? 0 : (vstup) > 1 ? 1 : (vstup))
#define INTERPOLUJ(v1, v2, gradient) ((v1) + ((v2) - (v1)) * KLAMP(gradient))

static double zBuf[ROZX640*ROZY480];

static void cshiara(VRCHOL *v1, VRCHOL *v2, unsigned long *buf,
                    int shiirka, int vyyshka, unsigned int farba)
{
    int x, y;

    if (v1->y > v2->y)
    {
        VRCHOL *vD = v1;
        v1 = v2;
        v2 = vD;
    }

    if (v1->y == v2->y)
    {
        for (x = min(v1->x, v2->x); x <= max(v1->x, v2->x); x++)
        {
            buf[x + ((int)(v1->y)) * shiirka] = farba;
        }

        return;
    }

    for (y = v1->y; y <= (v2->y); y++)
    {
        double gradient1 = (v1->y != (v2->y)
                            ? (y - v1->y) / (v2->y - v1->y) : 1);
        double z = INTERPOLUJ(v1->z, v2->z, gradient1);
        x = INTERPOLUJ(v1->x, v2->x, gradient1);

        if (z > zBuf[x + y * shiirka]) continue;

        zBuf[x + y * shiirka] = z;
        buf[x + y * shiirka] = farba;
    }
}

/* Najviac cshasu stojii taato funkcia. */
static void skenliinia(int y, VRCHOL *va, VRCHOL *vb, VRCHOL *vc, VRCHOL *vd,
                       unsigned long *buf, int shiirka, unsigned int farba)
{
    int x;

    double gradient1 = va->y != (vb->y) ? (y - va->y) / (vb->y - va->y) : 1;
    double gradient2 = vc->y != (vd->y) ? (y - vc->y) / (vd->y - vc->y) : 1;

    int sx = INTERPOLUJ(va->x, vb->x, gradient1);
    int kx = INTERPOLUJ(vc->x, vd->x, gradient2);

    double sz = INTERPOLUJ(va->z, vb->z, gradient1);
    double kz = INTERPOLUJ(vc->z, vd->z, gradient2);

    /* 2D rezanie. */
    if (y < 0 || y >= ROZY480) return;
    if (sx < 0) sx = 0;
    if (kx > ROZX640) kx = ROZX640;

    for (x = sx; x < kx; x++)
    {
        double gradient = ((double)(x - sx)) / ((double)(kx - sx));
        double z = INTERPOLUJ(sz, kz, gradient);

        if (z > zBuf[x + y * shiirka]) continue;

        zBuf[x + y * shiirka] = z;
        buf[x + y * shiirka] = farba;
    }
}

static void trojuhKresli(VRCHOL *v1, VRCHOL *v2, VRCHOL *v3,
                         unsigned long *buf, int shiirka, int vyyshka,
                         unsigned int farba)
{
    double smernicaV1V2, smernicaV1V3;
    /* Zoradenie vrcholov:
     * v1 je hore (najmenshie Y),
     * v3 je na spodku (najvaecshshie Y),
     * v2 je v strede. */
    if (v1->y > v2->y)
    {
        VRCHOL *vD = v1;
        v1 = v2;
        v2 = vD;
    }

    if (v2->y > v3->y)
    {
        VRCHOL *vD = v2;
        v2 = v3;
        v3 = vD;
    }

    if (v1->y > v2->y)
    {
        VRCHOL *vD = v1;
        v1 = v2;
        v2 = vD;
    }

    /* Probleem robiilo:
     * v2 ==== v1
     *   =     =
     *    =    =
     *     =   =
     *      =  =
     *        v3
     *
     * Rieshenie je zabezpecshit, aby v2 bolo vlhavo od v1,
     * a naasledne pouzhith vyynimku pri urcshovanii,
     * aby to vyshlo ako druhyy typ.
     */
    if (v1->y == v2->y && v1->x < v2->x)
    {
        VRCHOL *vD = v1;
        v1 = v2;
        v2 = vD;
    }

    /* Vyypocshet prevraatenyych smerniic (1 / smernica). */
    if ((v2->y) - (v1->y) > 0)
    {
        smernicaV1V2 = ((v2->x) - (v1->x)) / ((v2->y) - (v1->y));
    }
    else
    {
        smernicaV1V2 = 0;
    }

    if ((v3->y) - (v1->y) > 0)
    {
        smernicaV1V3 = ((v3->x) - (v1->x)) / ((v3->y) - (v1->y));
    }
    else
    {
        smernicaV1V3 = 0;
    }

    /*
     * v1
     * = =
     * =  =
     * =   v2
     * =  =
     * = =
     * v3
     */
    if (smernicaV1V2 > smernicaV1V3 && (v1->y != v2->y))
    {
        int y;

        for (y = v1->y; y <= v3->y; y++)
        {
            if (y < v2->y)
            {
                skenliinia(y, v1, v3, v1, v2, buf, shiirka, farba);
            }
            else
            {
                skenliinia(y, v1, v3, v2, v3, buf, shiirka, farba);
            }
        }
    }
    /*
     *     v1
     *    = =
     *   =  =
     * v2   =
     *   =  =
     *    = =
     *     v3
     */
    else
    {
        int y;

        for (y = v1->y; y <= v3->y; y++)
        {
            if (y < v2->y)
            {
                skenliinia(y, v1, v2, v1, v3, buf, shiirka, farba);
            }
            else
            {
                skenliinia(y, v2, v3, v1, v3, buf, shiirka, farba);
            }
        }
    }

    if (hrany)
    {
        cshiara(v1, v2, buf, shiirka, vyyshka, 0);
        cshiara(v1, v3, buf, shiirka, vyyshka, 0);
        cshiara(v3, v2, buf, shiirka, vyyshka, 0);
    }
}

static void trojProjekcia(KAMERA *kamera, TROJUH *trojuh, TROJUH *trojuh2)
{
    projekcia(kamera, &(trojuh->vrcholy[0]), &(trojuh2->vrcholy[0]));
    projekcia(kamera, &(trojuh->vrcholy[1]), &(trojuh2->vrcholy[1]));
    projekcia(kamera, &(trojuh->vrcholy[2]), &(trojuh2->vrcholy[2]));
    trojuh2->farba = trojuh->farba;
}

typedef struct {
    VEKTOR vektor;
    double vzdialenosth;
} ROVINA;

static double vzdialenosthOdRoviny(VRCHOL *vrchol, ROVINA *rovina)
{
    return (rovina->vektor.x * vrchol->x
            + rovina->vektor.y * vrchol->y
            + rovina->vektor.z * vrchol->z
            - rovina->vzdialenosth);
}

static void genBliizkaRovina(ROVINA *rovina)
{
    VEKTOR vektor = {0, 0, 1};

    rovina->vektor = vektor;
    rovina->vzdialenosth = 10;
}

static void genPostranneeRoviny(ROVINA *roviny, KAMERA *kamera)
{
    double horiz_uhol = atan2(ROZX640, kamera->div.z) - 0.1;
    double vert_uhol = atan2(ROZY480, kamera->div.z) - 0.1;
    VEKTOR vektor;

    /* Hornaa. */
    vektor.x = 0;
    vektor.y = cos(vert_uhol);
    vektor.z = sin(vert_uhol);

    roviny[0].vektor = vektor;
    roviny[0].vzdialenosth = 0;

    /* Lhavaa. */
    vektor.x = cos(horiz_uhol);
    vektor.y = 0;
    vektor.z = sin(horiz_uhol);

    roviny[1].vektor = vektor;
    roviny[1].vzdialenosth = 0;

    /* Pravaa. */
    vektor.x = -cos(horiz_uhol);
    vektor.y = 0;
    vektor.z = sin(horiz_uhol);

    roviny[2].vektor = vektor;
    roviny[2].vzdialenosth = 0;

    /* Dolnaa. */
    vektor.x = 0;
    vektor.y = -cos(vert_uhol);
    vektor.z = sin(vert_uhol);

    roviny[3].vektor = vektor;
    roviny[3].vzdialenosth = 0;

}

static int rezh(TROJUH *trojuh, ROVINA *rovina)
{
    if (vzdialenosthOdRoviny(&(trojuh->vrcholy[0]), rovina) < 0
        || vzdialenosthOdRoviny(&(trojuh->vrcholy[1]), rovina) < 0
        || vzdialenosthOdRoviny(&(trojuh->vrcholy[2]), rovina) < 0)
    {
        return (1);
    }

    return (0);
}

static void trojDelenie(KAMERA *kamera, TROJUH *trojuh)
{
    perspektiivneDelenie(kamera, &(trojuh->vrcholy[0]));
    perspektiivneDelenie(kamera, &(trojuh->vrcholy[1]));
    perspektiivneDelenie(kamera, &(trojuh->vrcholy[2]));
}

void zobrazTrojuhy(KAMERA *kamera,
                   void *sniimka, int shiirka, int vyyshka,
                   TROJUH *trojuhy, int pocshet)
{
    int i, j;
    TROJUH trojuh2;
    ROVINA roviny[5]; /* Bliizka, hornaa, lhavaa, pravaa, dolnaa. */

    kameraPredpocshiitaj(kamera);
    genBliizkaRovina(roviny);
    genPostranneeRoviny(roviny + 1, kamera);

    for (i = 0; i < pocshet; i++)
    {
        int p = 0;

        trojProjekcia(kamera, &(trojuhy[i]), &trojuh2);

        for (j = 0; j < 5; j++)
        {
            if (rezh(&trojuh2, roviny + j) == 1)
            {
                p = 1;
                break;
            }
        }
        if (p == 1) continue;

        trojDelenie(kamera, &trojuh2);
        trojuhKresli(&(trojuh2.vrcholy[2]),
                     &(trojuh2.vrcholy[1]),
                     &(trojuh2.vrcholy[0]),
                     sniimka, shiirka, vyyshka,
                     trojuh2.farba);
    }
}

void troj3dSniimka(void *sniimka, int shiirka, int vyyshka)
{
    unsigned long *buf = sniimka;
    int i;

    for (i = 0; i < ROZX640 * ROZY480; i++)
    {
        buf[i] = 0x00ffffff;
    }
    for (i = 0; i < ROZX640 * ROZY480; i++)
    {
        zBuf[i] = 9999.0;
    }
}
